@@ -317,6 +317,7 @@ sub hostname {
 
 sub server_setting_is_readable {
     my ($class, $key) = @_;
+    return 1 if $key eq 'fsck_checksum';
     return 0 if $key =~ /^fsck_/;
     return 1;
 }
@@ -387,6 +388,16 @@ sub server_setting_is_writable {
     # should probably restrict to (\d+)
     if ($key =~ /^queue_/) { return $any };
 
+    if ($key eq "fsck_checksum") {
+        return sub {
+            my $v = shift;
+            return "off" if $v eq "off";
+            return undef if $v eq "class";
+            return $v if MogileFS::Checksum->valid_alg($v);
+            die "Not a valid checksum algorithm";
+        }
+    }
+
     return 0;
 }
 
