@@ -4,6 +4,7 @@ use strict;
 use base 'MogileFS::Worker';
 use fields (
             'opt_nostat',          # bool: do we trust mogstoreds? skipping size stats?
+            'opt_checksum',        # (class|off|MD5) checksum mode
             );
 use MogileFS::Util qw(every error debug);
 use MogileFS::Config;
@@ -25,6 +26,9 @@ use constant EV_START_SEARCH     => "SRCH";
 use constant EV_FOUND_FID        => "FOND";
 use constant EV_RE_REPLICATE     => "REPL";
 use constant EV_BAD_COUNT        => "BCNT";
+use constant EV_BAD_CHECKSUM     => "BSUM";
+use constant EV_NO_CHECKSUM      => "NSUM";
+use constant EV_MULTI_CHECKSUM   => "MSUM";
 
 use POSIX ();
 
@@ -64,6 +68,12 @@ sub work {
         return unless @fids;
 
         $self->{opt_nostat} = MogileFS::Config->server_setting('fsck_opt_policy_only')     || 0;
+        my $alg = MogileFS::Config->server_setting_cached("fsck_checksum");
+        if (defined($alg) && $alg eq "off") {
+            $self->{opt_checksum} = "off";
+        } else {
+            $self->{opt_checksum} = MogileFS::Checksum->valid_alg($alg) ? $alg : 0;
+        }
         MogileFS::FID->mass_load_devids(@fids);
 
         # don't sleep in loop, next round, since we found stuff to work on
@@ -136,11 +146,20 @@ sub check_fid {
         return $fix->();
     }
 
+    # missing checksum row
+    if ($fid->class->hashtype && ! $fid->checksum) {
+        return $fix->();
+    }
+
     # in the fast case, do nothing else (don't check if assumed file
     # locations are actually there).  in the fast case, all we do is
     # check the replication policy, which is already done, so finish.
     return HANDLED if $self->{opt_nostat};
 
+    if ($self->{opt_checksum} && $self->{opt_checksum} ne "off") {
+        return $fix->();
+    }
+
     # stat each device to see if it's still there.  on first problem,
     # stop and go into the slow(er) fix function.
     my $err;
@@ -165,7 +184,8 @@ sub check_fid {
     });
 
     if ($rv) {
-        return HANDLED;
+        return ($fid->class->hashtype && !($self->{opt_checksum} && $self->{opt_checksum} eq "off"))
+            ? $fix->() : HANDLED;
     } elsif ($err eq "stalled") {
         return STALLED;
     } elsif ($err eq "needfix") {
@@ -208,6 +228,9 @@ sub fix_fid {
     my @good_devs;
     my @bad_devs;
     my %already_checked;  # devid -> 1.
+    my $alg = $fid->class->hashname || $self->{opt_checksum};
+    my $checksums = {};
+    my $ping_cb = sub { $self->still_alive };
 
     my $check_dfids = sub {
         my $is_desperate_mode = shift;
@@ -227,6 +250,23 @@ sub fix_fid {
             die "dev " . $dev->id . " unreachable" unless defined $disk_size;
 
             if ($disk_size == $fid->length) {
+                if ($alg && $alg ne "off") {
+                    my $digest = $self->checksum_on_disk($dfid, $alg, $ping_cb);
+                    unless (defined $digest) {
+                        die "dev " . $dev->id . " unreachable";
+                    }
+
+                    # DELETE could've hit right after size check
+                    if ($digest eq "-1") {
+                        unless ($is_desperate_mode) {
+                            $fid->fsck_log(EV_FILE_MISSING, $dev);
+                        }
+                        push @bad_devs, $dfid->device;
+                        next;
+                    }
+                    push @{$checksums->{$digest} ||= []}, $dfid->device;
+                }
+
                 push @good_devs, $dfid->device;
                 # if we were doing a desperate search, one is enough, we can stop now!
                 return if $is_desperate_mode;
@@ -282,6 +322,8 @@ sub fix_fid {
     # in case the devcount or similar was fixed.
     $fid->want_reload;
 
+    $self->fix_checksums($fid, $checksums) if $alg && $alg ne "off";
+
     # Note: this will reload devids, if they called 'note_on_device'
     # or 'forget_about_device'
     unless ($fid->devids_meet_policy) {
@@ -299,6 +341,16 @@ sub fix_fid {
     return HANDLED;
 }
 
+sub forget_file_on_with_bad_checksums {
+    my ($self, $fid, $checksums) = @_;
+    foreach my $bdevs (values %$checksums) {
+        foreach my $bdev (@$bdevs) {
+            error("removing file_on mapping for fid=" . $fid->id . ", dev=" . $bdev->id);
+            $fid->forget_about_device($bdev);
+        }
+    }
+}
+
 # returns 0 on missing,
 # undef on connectivity error,
 # else size of file on disk (after HTTP HEAD or mogstored stat)
@@ -308,6 +360,96 @@ sub size_on_disk {
     return $dfid->size_on_disk;
 }
 
+# returns -1 on missing,
+# undef on connectivity error,
+# else checksum of file on disk (after HTTP GET or mogstored read)
+sub checksum_on_disk {
+    my ($self, $dfid, $alg, $ping_cb) = @_;
+    return undef if $dfid->device->dstate->is_perm_dead;
+    return $dfid->checksum_on_disk($alg, $ping_cb, "fsck");
+}
+
+sub bad_checksums_errmsg {
+    my ($self, $alg, $checksums) = @_;
+    my @err;
+
+    foreach my $checksum (keys %$checksums) {
+        my $bdevs = join(",", map { $_->id } @{$checksums->{$checksum}});
+        $checksum = unpack("H*", $checksum);
+        push @err, "$alg:$checksum on devids=[$bdevs]"
+    }
+
+    return join('; ', @err);
+}
+
+# we don't now what checksum the file is supposed to be, but some
+# of the devices had checksums that didn't match the other(s).
+sub auto_checksums_bad {
+    my ($self, $fid, $checksums) = @_;
+    my $alg = $self->{opt_checksum};
+    my $err = $self->bad_checksums_errmsg($alg, $checksums);
+
+    error("$fid has multiple checksums: $err");
+    $fid->fsck_log(EV_MULTI_CHECKSUM);
+}
+
+sub all_checksums_bad {
+    my ($self, $fid, $checksums) = @_;
+    my $alg = $fid->class->hashname or return; # class could've changed
+    my $cur_checksum = $fid->checksum;
+    my $err = $self->bad_checksums_errmsg($alg, $checksums);
+    my $cur = $cur_checksum ? "Expected: $cur_checksum"
+                            : "No known valid checksum";
+    error("all checksums bad: $err. $cur");
+    $fid->fsck_log(EV_BAD_CHECKSUM);
+}
+
+sub fix_checksums {
+    my ($self, $fid, $checksums) = @_;
+    my $cur_checksum = $fid->checksum;
+    my @all_checksums = keys(%$checksums);
+
+    if (scalar(@all_checksums) == 1) { # all checksums match, good!
+        my $disk_checksum = $all_checksums[0];
+        if ($cur_checksum) {
+            if ($cur_checksum->{checksum} ne $disk_checksum) {
+                $fid->fsck_log(EV_BAD_CHECKSUM);
+            }
+        } else { # fresh row to checksum
+            my $hashtype = $fid->class->hashtype;
+
+            # we store this in the database
+            if ($hashtype) {
+                my %row = (
+                    fid => $fid->id,
+                    checksum => $disk_checksum,
+                    hashtype => $hashtype,
+                );
+                my $new_checksum = MogileFS::Checksum->new(\%row);
+                debug("creating new checksum=$new_checksum");
+                $fid->fsck_log(EV_NO_CHECKSUM);
+                $new_checksum->save;
+            } else {
+                my $hex_checksum = unpack("H*", $disk_checksum);
+                my $alg = $self->{opt_checksum};
+                debug("fsck_checksum=auto good: $fid $alg:$hex_checksum");
+            }
+        }
+    } elsif ($cur_checksum) {
+        my $good = delete($checksums->{$cur_checksum->{checksum}});
+        if ($good && (scalar(@$good) > 0)) {
+            $self->forget_file_on_with_bad_checksums($fid, $checksums);
+            # will fail $fid->devids_meet_policy and re-replicate
+        } else {
+            $self->all_checksums_bad($fid, $checksums);
+        }
+    } elsif ($self->{opt_checksum}) {
+        $self->auto_checksums_bad($fid, $checksums);
+    } else {
+        $self->all_checksums_bad($fid, $checksums);
+    }
+}
+
 1;
 
 # Local Variables:
