@@ -355,6 +355,12 @@ sub cmd_create_close {
     my $fidid = $args->{fid}    or return $self->err_line("no_fid");
     my $devid = $args->{devid}  or return $self->err_line("no_devid");
     my $path  = $args->{path}   or return $self->err_line("no_path");
+    my $checksum = $args->{checksum};
+
+    if ($checksum) {
+        $checksum = eval { MogileFS::Checksum->from_string($fidid, $checksum) };
+        return $self->err_line("invalid_checksum_format") if $@;
+    }
 
     my $fid  = MogileFS::FID->new($fidid);
     my $dfid = MogileFS::DevFID->new($devid, $fid);
@@ -391,7 +397,8 @@ sub cmd_create_close {
     }
 
     # get size of file and verify that it matches what we were given, if anything
-    my $size = MogileFS::HTTPFile->at($path)->size;
+    my $httpfile = MogileFS::HTTPFile->at($path);
+    my $size = $httpfile->size;
 
     # size check is optional? Needs to support zero byte files.
     $args->{size} = -1 unless $args->{size};
@@ -408,6 +415,20 @@ sub cmd_create_close {
         return $self->err_line("size_mismatch", "Expected: $args->{size}; actual: $size; path: $path")
     }
 
+    # checksum validation is optional as it can be very expensive
+    # However, we /always/ verify it if the client wants us to, even
+    # if the class does not enforce or store it.
+    if ($checksum && $args->{checksumverify}) {
+        my $alg = $checksum->hashname;
+        my $actual = $httpfile->digest($alg, sub { $self->still_alive });
+        if ($actual ne $checksum->{checksum}) {
+            $failed->();
+            $actual = "$alg:" . unpack("H*", $actual);
+            return $self->err_line("checksum_mismatch",
+                           "Expected: $checksum; actual: $actual; path: $path");
+        }
+    }
+
     # see if we have a fid for this key already
     my $old_fid = MogileFS::FID->new_from_dmid_and_key($dmid, $key);
     if ($old_fid) {
@@ -424,6 +445,8 @@ sub cmd_create_close {
     # insert file_on row
     $dfid->add_to_db;
 
+    $checksum->maybe_save($dmid, $trow->{classid}) if $checksum;
+
     $sto->replace_into_file(
                             fidid   => $fidid,
                             dmid    => $dmid,
@@ -561,6 +584,15 @@ sub cmd_file_debug {
     }
     $ret->{devids} = join(',', @devids) if @devids;
 
+    # Always look for a checksum
+    my $checksum = Mgd::get_store()->get_checksum($fidid);
+    if ($checksum) {
+        $checksum = MogileFS::Checksum->new($checksum);
+        $ret->{checksum} = $checksum->info;
+    } else {
+        $ret->{checksum} = 'NONE';
+    }
+
     # Return file row (if found) and all other data.
     my %toret = (fid => $fid, tempfile => $tfile, replqueue => $repl,
         delqueue => $del, rebqueue => $reb, fsckqueue => $fsck);
@@ -600,8 +632,17 @@ sub cmd_file_info {
     my $ret = {};
     $ret->{fid}      = $fid->id;
     $ret->{domain}   = Mgd::domain_factory()->get_by_id($fid->dmid)->name;
-    $ret->{class}    = Mgd::class_factory()->get_by_id($fid->dmid,
-        $fid->classid)->name;
+    my $class = Mgd::class_factory()->get_by_id($fid->dmid, $fid->classid);
+    $ret->{class}    = $class->name;
+    if ($class->{hashtype}) {
+        my $checksum = Mgd::get_store()->get_checksum($fid->id);
+        if ($checksum) {
+            $checksum = MogileFS::Checksum->new($checksum);
+            $ret->{checksum} = $checksum->info;
+        } else {
+            $ret->{checksum} = "MISSING";
+        }
+    }
     $ret->{key}      = $key;
     $ret->{'length'} = $fid->length;
     $ret->{devcount} = $fid->devcount;
@@ -840,6 +881,13 @@ sub cmd_create_class {
         return $self->err_line('invalid_replpolicy', $@) if $@;
     }
 
+    my $hashtype = $args->{hashtype};
+    if ($hashtype && $hashtype ne 'NONE') {
+        my $tmp = $MogileFS::Checksum::NAME2TYPE{$hashtype};
+        return $self->err_line('invalid_hashtype') unless $tmp;
+        $hashtype = $tmp;
+    }
+
     my $sto = Mgd::get_store();
     my $dmid  = $sto->get_domainid_by_name($domain) or
         return $self->err_line('domain_not_found');
@@ -866,6 +914,10 @@ sub cmd_create_class {
     # don't erase an existing replpolicy if we're not setting a new one.
     $sto->update_class_replpolicy(dmid => $dmid, classid => $clsid,
         replpolicy => $replpolicy) if $replpolicy;
+    if ($hashtype) {
+        $sto->update_class_hashtype(dmid => $dmid, classid => $clsid,
+            hashtype => $hashtype eq 'NONE' ? undef : $hashtype);
+    }
 
     # return success
     return $self->cmd_clear_cache({ class => $class, mindevcount => $mindevcount, domain => $domain });
@@ -990,6 +1042,7 @@ sub cmd_get_domains {
             $ret->{"domain${dm_n}class${cl_n}mindevcount"} = $cl->mindevcount;
             $ret->{"domain${dm_n}class${cl_n}replpolicy"}  =
                 $cl->repl_policy_string;
+            $ret->{"domain${dm_n}class${cl_n}hashtype"} = $cl->hashtype_string;
         }
         $ret->{"domain${dm_n}classes"} = $cl_n;
     }
@@ -1426,7 +1479,8 @@ sub cmd_fsck_start {
     my $intss       = sub { MogileFS::Config->server_setting($_[0]) || 0 };
     my $checked_fid = $intss->("fsck_highest_fid_checked");
     my $final_fid   = $intss->("fsck_fid_at_end");
-    if ($checked_fid && $final_fid && $checked_fid >= $final_fid) {
+    if (($checked_fid && $final_fid && $checked_fid >= $final_fid) ||
+        (!$final_fid && !$checked_fid)) {
         $self->_do_fsck_reset or return $self->err_line;
     }
 
